"""
Utilities to help with catalogs.
"""

import fnmatch
import pathlib
import re

import cf_xarray  # noqa
import numpy as np
import pandas as pd
import requests
import yaml

from intake.catalog import Catalog
from siphon.catalog import TDSCatalog

import model_catalogs as mc


def astype(value, type_):
    """Return `value` as type `type_`.

    Particularly made to work correctly for returning string, `PosixPath`, or `Timestamp` as list.
    """
    if not isinstance(value, type_):
        if type_ == list and isinstance(
            value, (str, pathlib.PurePath, pd.Timestamp, Catalog)
        ):
            return [value]
        return type_(value)
    return value


def status(urlpath, suffix=".das"):
    """Check status of server for urlpath.

    Parameters
    ----------
    urlpath : str
        Path to file location to check.

    Returns
    -------
    bool
        If True, server was reachable.
    """

    resp = requests.get(urlpath + suffix)
    if resp.status_code != 200:
        status = False
    else:
        status = True
    return status


def file2dt(filename):
    """Return Timestamp of NOAA OFS filename

    ...without reading in the filename to xarray. See `docs <https://model-catalogs.readthedocs.io/en/latest/aggregations.html>`_ for details on the formula. Most NOAA OFS models have 1 timestep per file, but NYOFS has 6.

    Parameters
    ----------
    filename : str
        Filename for which to decipher datetime. Can be full path or just file name.

    Returns
    -------
    Timestamp
        pandas Timestamp of the time(s) in the file.

    Examples
    --------
    >>> url = 'https://www.ncei.noaa.gov/thredds/dodsC/model-cbofs-files/2022/07/nos.cbofs.fields.n001.20220701.t00z.nc'
    >>> mc.filename2datetime(url)
    Timestamp('2022-06-30 19:00:00')
    """

    # strip off path if present since can mess up the matching
    filename = filename.split("/")[-1]

    # read in date from filename
    date = pd.to_datetime(filename, format="%Y%m%d", exact=False)  # has time 00:00

    # pull timing cycle from filename
    regex = re.compile(".t[0-9]{2}z.")
    cycle = int(regex.findall(filename)[0][2:4])

    # NYOFS: multiple times per file
    if fnmatch.fnmatch(filename, "*.nowcast.*"):

        date = [date + pd.Timedelta(f"{cycle - dt} hours") for dt in range(6)[::-1]]

    # NYOFS: multiple times per file
    elif fnmatch.fnmatch(filename, "*.forecast.*"):

        # models all have different forecast lengths! Though only nyofs is left in this category
        if "nyofs" in filename:
            nfiles = 54

        date = [date + pd.Timedelta(f"{cycle + 1 + dt} hours") for dt in range(nfiles)]

    # Main style of NOAA OFS files, 1 file per time step
    elif fnmatch.fnmatch(filename, "*.n???.*") or fnmatch.fnmatch(filename, "*.f???.*"):

        # pull hours from filename
        regex = re.compile(".[n,f][0-9]{3}.")
        hour = int(regex.findall(filename)[0][2:-1])

        # calculate hours
        dt = cycle + hour

        # if nowcast file, subtract 6 hours
        if fnmatch.fnmatch(filename, "*.n???.*"):
            dt -= 6

        # construct datetime. dt might be negative.
        date += pd.Timedelta(f"{dt} hours")

    return date


def get_fresh_parameter(filename, source):
    """Get freshness parameter, based on the filename.

    A freshness parameter is stored in ``__init__`` for required scenarios which is looked up using the logic in this function, based on the filename. The source is checked for most types of actions for an overriding freshness parameter value, otherwise the default is used.

    Parameters
    ----------
    filename : Path
        Filename to determine freshness.
    source : Intake Source
        Source from which to check for an overriding freshness parameter. Is not used for "compiled" catalog files.

    Returns
    -------
    str
        mu, a pandas Timedelta-interpretable string describing the amount of time that filename should be considered fresh before needing to be recalculated.
    """

    # a start or end datetime file
    if filename.parent == mc.CACHE_PATH_AVAILABILITY:

        if source is None:
            raise ValueError("source cannot be None for this freshness calculation.")

        # which type are we after
        if "start" in filename.name:
            parameter = "start"
        elif "end" in filename.name:
            parameter = "end"
        elif "catrefs" in filename.name:
            parameter = "catrefs"

        # check for overriding freshness parameter in source metadata
        if "freshness" in source.metadata and parameter in source.metadata["freshness"]:
            mu = source.metadata["freshness"][parameter]
        else:
            mu = mc.FRESH[parameter]
            # a file of file locs for aggregation
    elif filename.parent == mc.CACHE_PATH_FILE_LOCS:

        if source is None:
            raise ValueError("source cannot be None for this freshness calculation.")

        parameter = "file_locs"
        # check for overriding freshness parameter in source metadata
        if "freshness" in source.metadata and parameter in source.metadata["freshness"]:
            mu = source.metadata["freshness"][parameter]
        else:
            mu = mc.FRESH[parameter]

    # a compiled catalog file
    elif filename.parent == mc.CACHE_PATH_COMPILED:
        mu = mc.FRESH["compiled"]

    return mu


def is_fresh(filename, source=None):
    """Check if file called filename is fresh.

    If filename doesn't exist, return False.

    Parameters
    ----------
    filename : Path
        Filename to determine freshness
    source : Intake Source
        Source from which to check for an overriding freshness parameter. Is not used for "compiled" catalog files.

    Returns
    -------
    Boolean
        True if fresh and False if not or if filename is not found.
    """

    now = pd.Timestamp.today(tz="UTC")
    try:
        filetime = pd.Timestamp(filename.stat().st_mtime_ns).tz_localize("UTC")

        mu = get_fresh_parameter(filename, source=source)

        return now - filetime < pd.Timedelta(mu)

    except FileNotFoundError:
        return False


def find_bbox(ds, dd=None, alpha=None):
    """Determine bounds and boundary of model.

    Parameters
    ----------
    ds: Dataset
        xarray Dataset containing model output.
    dd: int, optional
        Number to decimate model output lon/lat, as a stride.
    alpha: float, optional
        Number for alphashape to determine what counts as the convex hull. Larger number is more detailed, 1 is a good starting point.

    Returns
    -------
    List
        Contains the name of the longitude and latitude variables for ds, geographic bounding box of model output (`[min_lon, min_lat, max_lon, max_lat]`), low res and high res wkt representation of model boundary.
    """

    import shapely.geometry

    hasmask = False

    try:
        lon = ds.cf["longitude"].values
        lat = ds.cf["latitude"].values
        lonkey = ds.cf["longitude"].name
        latkey = ds.cf["latitude"].name

    except KeyError:
        if "lon_rho" in ds:
            lonkey = "lon_rho"
            latkey = "lat_rho"
        else:
            lonkey = list(ds.cf[["longitude"]].coords.keys())[0]
            # need to make sure latkey matches lonkey grid
            latkey = f"lat{lonkey[3:]}"
        # In case there are multiple grids, just take first one;
        # they are close enough
        lon = ds[lonkey].values
        lat = ds[latkey].values

    # check for corresponding mask (rectilinear and curvilinear grids)
    if any([var for var in ds.data_vars if "mask" in var]):
        if ("mask_rho" in ds) and (lonkey == "lon_rho"):
            maskkey = lonkey.replace("lon", "mask")
        elif "mask" in ds:
            maskkey = "mask"
        else:
            maskkey = None
        if maskkey in ds:
            lon = ds[lonkey].where(ds[maskkey] == 1).values
            lon = lon[~np.isnan(lon)].flatten()
            lat = ds[latkey].where(ds[maskkey] == 1).values
            lat = lat[~np.isnan(lat)].flatten()
            hasmask = True

    # This is structured, rectilinear
    # GFS, RTOFS, HYCOM
    if (lon.ndim == 1) and ("nele" not in ds.dims) and not hasmask:
        nlon, nlat = ds["lon"].size, ds["lat"].size
        lonb = np.concatenate(([lon[0]] * nlat, lon[:], [lon[-1]] * nlat, lon[::-1]))
        latb = np.concatenate((lat[:], [lat[-1]] * nlon, lat[::-1], [lat[0]] * nlon))
        # boundary = np.vstack((lonb, latb)).T
        p = shapely.geometry.Polygon(zip(lonb, latb))
        p0 = p.simplify(1)
        # Now using the more simplified version because all of these models are boxes
        p1 = p0

    elif hasmask or ("nele" in ds.dims):  # unstructured

        assertion = (
            "dd and alpha need to be defined in the catalog metadata for this model."
        )
        assert dd is not None and alpha is not None, assertion

        # need to calculate concave hull or alphashape of grid
        import alphashape

        # downsample a bit to save time, still should clearly see shape of domain
        lon, lat = lon[::dd], lat[::dd]
        pts = list(zip(lon, lat))

        # low res, same as convex hull
        p0 = alphashape.alphashape(pts, 0.0)
        p1 = alphashape.alphashape(pts, alpha)

    # useful things to look at: p.wkt  #shapely.geometry.mapping(p)
    return lonkey, latkey, list(p0.bounds), p1.wkt
    # return lonkey, latkey, list(p0.bounds), p0.wkt, p1.wkt


def filedates2df(filelocs):
    """Set up dataframe of datetimes to filenames.

    Parameters
    ----------
    filelocs : list of str
        File locations.

    Returns
    -------
    DataFrame
        Contains the index datetimes corresponding to file locations (column 'filenames').
    """

    # 1+ number of times possible from mc.file2dt, need the number of filenames to match
    filedates, filenames = [], []
    for fname in filelocs:
        filedate = mc.astype(mc.file2dt(fname), list)
        filenames.extend([fname] * len(filedate))
        filedates.extend(filedate)

    # Make dataframe
    df = pd.DataFrame(index=filedates, data={"filenames": filenames})

    # Sort resulting df by filenames and then by index which is the datetime of each file
    df = df.sort_values(axis="index", by="filenames").sort_index()

    # remove rows if index is duplicated, sorting makes it so nowcast files are kept
    df = df[~df.index.duplicated(keep="first")]

    return df


def agg_for_date(date, strings, filetype, is_forecast=False, pattern=None):
    """Select NOAA OFS-style nowcast/forecast files for aggregation.

    This function finds the files whose path includes the given date, regardless of times which might change the date forward or backward.

    Parameters
    ----------
    date: str of datetime, pd.Timestamp
        Date of day to find model output files for. Doesn't pay attention to hours/minutes seconds.
    strings: list
        List of strings to be filtered. Expected to be file locations from a thredds catalog.
    filetype: str
        Which filetype to use. Every NOAA OFS model has "fields" available, but some have "regulargrid" or "2ds" also. This availability information is in the catalog metadata for the model under `filetypes` metadata.
    is_forecast: bool, optional
        If True, then date is the last day of the time period being sought and the forecast files should be brought in along with the nowcast files, to get the model output the length of the forecast out in time. The forecast files brought in will have the latest timing cycle of the day that is available. If False, all nowcast files (for all timing cycles) are brought in.
    pattern: str, optional
        If a model file pattern doesn't match that assumed in this code, input one that will work. Currently only NYOFS doesn't match but the pattern is built into the catalog file.

    Returns
    -------
    List
        Contains URLs for where to find all of the model output files that match the keyword arguments. List is not sorted correctly for times (this happens later).
    """

    date = astype(date, pd.Timestamp)

    if pattern is None:
        pattern = date.strftime(f"*{filetype}*.n*.%Y%m%d.t??z.*")
    else:
        pattern = eval(f"f'{pattern}'")

    # if using forecast, find nowcast and forecast files for the latest full timing cycle
    if is_forecast:

        import re

        # Find the most recent, complete timing cycle for the forecast day
        regex = re.compile(".t[0-9]{2}z.")
        # substrings: list of repeated strings of hours, e.g. ['12', '06', '00', '12', ...]
        subs = [substr[2:4] for substr in regex.findall("".join(strings))]
        # unique str times, in increasing order, e.g. ['00', '06', '12']
        times = sorted(list(set(subs)))
        # choose the timing cycle that is latest but also has the most times available
        # sometimes the forecast files aren't available yet so don't want to use that time
        cycle = sorted(times, key=subs.count)[-1]  # noqa: F841

        # find all nowcast files with only timing "cycle"
        # replace '.t??z.' in pattern with '.t{cycle}z.' to get the latest timing cycle only
        pattern1 = pattern.replace(".t??z.", ".t{cycle}z.").replace(".n*.", ".[n,f]*.")
        pattern1 = eval(f"f'{pattern1}'")  # use `eval` to sub in value of `cycle`
        # sort nowcast files alone to get correct time order
        fnames_cycle = fnmatch.filter(strings, pattern1)

        # Include the nowcast files between the start of the day and when the time series
        # represented in fnames begins
        # fnames_now = find_nowcast_cycles(strings, pattern)
        fnames_nowcast = fnmatch.filter(strings, pattern)

        fnames = fnames_cycle + fnames_nowcast  # combine

        if len(fnames) == 0:
            raise ValueError(
                f"Error finding filenames. Filenames found so far: {fnames}. "
                "Maybe you have the wrong source for the days requested."
            )

    # if not using forecast, find all nowcast files matching pattern
    else:
        fnames = fnmatch.filter(strings, pattern)

    return fnames


def find_catrefs(catloc):
    """Find hierarchy of catalog references for thredds catalog.

    Parameters
    ----------
    catloc: str
        Search in thredds catalog structure from base catalog, catloc.

    Returns
    -------
    list
        Contains tuples containing the hierarchy of directories in the thredds catalog structure to get to where the datafiles start.
    """

    # 0th level catalog
    cat = TDSCatalog(catloc)
    catrefs_to_check = list(cat.catalog_refs)
    # only keep numerical directories
    catrefs_to_check = [catref for catref in catrefs_to_check if catref.isnumeric()]

    # 1st level catalog
    catrefs_to_check1 = [
        cat.catalog_refs[catref].follow().catalog_refs for catref in catrefs_to_check
    ]

    # Combine the catalog references together
    catrefs = [
        (catref0, catref1)
        for catref0, catrefs1 in zip(catrefs_to_check, catrefs_to_check1)
        for catref1 in catrefs1
    ]

    # Check first one to see if there are more catalog references or not
    cat_ref_test = (
        cat.catalog_refs[catrefs[0][0]]
        .follow()
        .catalog_refs[catrefs[0][1]]
        .follow()
        .catalog_refs
    )
    # If there are more catalog references, run another level of catalog and combine, ## 2
    if len(cat_ref_test) > 0:
        catrefs2 = [
            cat.catalog_refs[catref[0]]
            .follow()
            .catalog_refs[catref[1]]
            .follow()
            .catalog_refs
            for catref in catrefs
        ]
        catrefs = [
            (catref[0], catref[1], catref22)
            for catref, catref2 in zip(catrefs, catrefs2)
            for catref22 in catref2
        ]

    return catrefs


def find_filelocs(catref, catloc, filetype="fields"):
    """Find thredds file locations.

    Parameters
    ----------
    catref: tuple
        2 or 3 labels describing the directories from catlog to get the data
        locations.
    catloc: str
        Base thredds catalog location.
    filetype: str
        Which filetype to use. Every NOAA OFS model has "fields" available, but
        some have "regulargrid" or "2ds" also (listed in separate catalogs in the
        model name).

    Returns
    -------
    list
        Locations of files found from catloc to hierarchical location described by catref.
    """

    filelocs = []
    cat = TDSCatalog(catloc)
    if len(catref) == 2:
        catref1, catref2 = catref
        cat1 = cat.catalog_refs[catref1].follow()
        cat2 = cat1.catalog_refs[catref2].follow()
        last_cat = cat2
        datasets = cat2.datasets
    elif len(catref) == 3:
        catref1, catref2, catref3 = catref
        cat1 = cat.catalog_refs[catref1].follow()
        cat2 = cat1.catalog_refs[catref2].follow()
        cat3 = cat2.catalog_refs[catref3].follow()
        last_cat = cat3
        datasets = cat3.datasets

    for dataset in datasets:
        if (
            "stations" not in dataset
            and "vibrioprob" not in dataset
            and filetype in dataset
        ):

            url = last_cat.datasets[dataset].access_urls["OPENDAP"]
            filelocs.append(url)
    return filelocs


def calculate_boundaries(cats, save_files=True, return_boundaries=False):
    """Calculate boundary information for all models.

    This loops over all input catalogs and will try with multiple model_source if necessary (in case servers aren't working) to access the example model output files and calculate the bounding box and numerical domain boundary. The numerical domain boundary is calculated using `alpha_shape` with previously-chosen parameters stored in the original model catalog files. The bounding box and boundary string representation (as WKT) are then saved to files.

    The files are saved the first time you run this function, so this function should only be rerun if you suspect that a model domain has changed or you have a new model catalog.

    Parameters
    ----------
    cats : Catalog, list of Catalogs
        The Catalog or Catalogs for which to find boundaries.
    save_files : boolean, optional
        Whether to save files or not. Defaults to True. Saves to ``mc.FILE_PATH_BOUNDARIES(cat_loc.name)``.
    return_boundaries : boolean, optional
        Whether to return boundaries information from this call. Defaults to False.

    Examples
    --------

    Calculate boundary information for CBOFS:

    >>> import model_catalogs as mc
    >>> main_cat = mc.setup()
    >>> mc.calculate_boundaries(main_cat["CBOFS"])
    """

    # loop over all orig catalogs
    boundaries = {}
    for cat in mc.astype(cats, list):

        # loop over available sources and use the first that works
        for model_source in list(cat):
            source_orig = cat[model_source]
            source_transform = mc.transform_source(source_orig)

            # need to make catalog to transfer information properly from
            # source_orig to source_transform
            cat_transform = mc.make_catalog(
                source_transform,
                full_cat_name=cat.name,  # model name
                full_cat_description=cat.description,
                full_cat_metadata=cat.metadata,
                cat_driver=mc.process.DatasetTransform,
                cat_path=None,
                save_catalog=False,
            )

            if cat_transform[model_source].status:

                # read in model output
                ds = cat_transform[model_source].to_dask()
                break

        # find boundary information for model
        if "alpha_shape" in cat.metadata:
            dd, alpha = cat.metadata["alpha_shape"]
        else:
            dd, alpha = None, None
        lonkey, latkey, bbox, wkt = mc.find_bbox(ds, dd=dd, alpha=alpha)

        ds.close()

        # save boundary info to file
        if save_files:
            with open(mc.FILE_PATH_BOUNDARIES(cat.name.lower()), "w") as outfile:
                yaml.dump({"bbox": bbox, "wkt": wkt}, outfile, default_flow_style=False)

        if return_boundaries:
            boundaries[cat.name] = {"bbox": bbox, "wkt": wkt}

    if return_boundaries:
        return boundaries



"""
Everything dealing with the catalogs.
"""

import warnings

from datetime import datetime
from pathlib import Path, PurePath

import cf_xarray  # noqa
import intake
import intake.source.derived
import pandas as pd
import requests
import yaml

from datetimerange import DateTimeRange
from dateutil.parser import parse
from intake.catalog import Catalog
from intake.catalog.local import LocalCatalogEntry
from intake_xarray.opendap import OpenDapSource

import model_catalogs as mc

from model_catalogs.process import DatasetTransform


DEFAULT = datetime(1970, 1, 1, 22, 22, 22)


def make_catalog(
    cats,
    full_cat_name,
    full_cat_description,
    full_cat_metadata,
    cat_driver,
    cat_path=None,
    save_catalog=True,
    return_cat: bool = True,
):
    """Construct single catalog from multiple catalogs or sources.

    Parameters
    ----------
    cats: list
       List of Intake catalog or source objects that will be combined into a single catalog.
    full_cat_name: str
       Name of overall catalog.
    full_cat_descrption: str
       Description of overall catalog.
    full_cat_metadata: dict
       Dictionary of metadata for overall catalog.
    cat_driver: str or Intake object or list
       Driver to apply to all catalog entries. For example:

       * `intake.catalog.local.YAMLFileCatalog`
       * `'opendap'`

       If list, must be same length as cats and contains drivers that
       correspond to cats.
    cat_path: Path object, optional
       Path with catalog name to use for saving catalog. With or without yaml suffix. If not provided,
       will use `full_cat_name`.
    save_catalog : bool, optional
        Defaults to True, and saves to cat_path.
    return_cat : bool, optional
        Return catalog.

    Returns
    -------
    Intake Catalog
        A single catalog made from multiple catalogs or sources.

    Examples
    --------

    Make catalog:

    >>> make_catalog([list of Intake sources or catalogs], 'catalog name', 'catalog desc',
                     {}, 'opendap', save_catalog=False)
    """

    if cat_path is None and save_catalog:
        cat_path = Path(full_cat_name)
    # if save_catalog:
    #     cat_path = cat_path / full_cat_name.lower()
    #     # cat_path = f"{cat_path}/{full_cat_name.lower()}"
    # if save_catalog and ("yaml" not in str(cat_path)) and ("yml" not in str(cat_path)):
    #     cat_path = cat_path.with_suffix(".yaml")

    if not isinstance(cats, list):
        cats = [cats]
    if not isinstance(cat_driver, list):
        cat_driver = [cat_driver] * len(cats)
    assert len(cat_driver) == len(
        cats
    ), "Number of catalogs and catalog drivers must match"

    # create dictionary of catalog entries
    entries = {
        cat.name: LocalCatalogEntry(
            cat.name.upper() if isinstance(cat, Catalog) else cat.name,
            description=cat.description,
            driver=catd,
            args=cat._yaml()["sources"][cat.name]["args"],
            metadata=cat.metadata,
        )
        for cat, catd in zip(cats, cat_driver)
    }

    # create catalog
    cat = Catalog.from_dict(
        entries,
        name=full_cat_name,
        description=full_cat_description,
        metadata=full_cat_metadata,
    )

    # save catalog
    if save_catalog:
        cat.save(cat_path.with_suffix(".yaml"))

    if return_cat:
        return cat


def open_catalog(
    cat_loc,
    return_cat=True,
    save_catalog=False,
    override=False,
    boundaries=False,
    save_boundaries=False,
):
    """Open an intake catalog file and set up code to apply processing/transform.

    Optionally calculate the boundaries of the model represented in cat_log.

    Note that saved boundaries files will be saved under the name inside the catalog, not the name of the file if you input a catalog path.

    Parameters
    ----------
    cat_loc : str, Catalog
        The catalog to open. cat_loc can be the representation of a path to a catalog file (string or Path) or it can be a Catalog object.
    return_cat : bool, optional
        Return catalog from function. Defaults to True.
    save_catalog : bool, optional
        Defaults to False, and saves to mc.CACHE_PATH_COMPILED(model).
    override : boolean, optional
        Use `override=True` to calculate boundaries of the model regardless of whether the file already exists.
    boundaries : boolean, optional
        If True, find previously-saved or calculate domain boundary of model.
    save_boundaries : bool, optional
        Defaults to False, and saves to mc.FILE_PATH_BOUNDARIES(model).
    """

    if isinstance(cat_loc, Catalog):
        cat_orig = cat_loc
    else:
        cat_loc = PurePath(cat_loc)
        cat_orig = intake.open_catalog(cat_loc)

    if boundaries:
        if mc.FILE_PATH_BOUNDARIES(cat_orig.name.lower()).is_file() and not override:
            # add previously-saved boundary info
            # this was calculated with mc.calculate_boundaries()
            with open(mc.FILE_PATH_BOUNDARIES(cat_orig.name.lower()), "r") as stream:
                boundary = yaml.safe_load(stream)
        else:
            boundary = mc.calculate_boundaries(
                cat_orig, save_files=save_boundaries, return_boundaries=True
            )[cat_orig.name]

        # add to cat_orig metadata
        cat_orig.metadata["bounding_box"] = boundary["bbox"]
        cat_orig.metadata["geospatial_bounds"] = boundary["wkt"]

    # get transform of each original catalog file, which points to
    # original file but applies metadata from original catalog file
    # to the resulting dataset after calling `to_dask()`
    source_transforms = [
        mc.transform_source(cat_orig[model_source]) for model_source in list(cat_orig)
    ]

    metadata = cat_orig.metadata
    metadata.update({"cat_path": cat_orig.path})

    # need to make catalog to transfer information properly from
    # source_orig to source_transform
    cat = mc.make_catalog(
        source_transforms,
        full_cat_name=cat_orig.name,  # model name
        full_cat_description=cat_orig.description,
        full_cat_metadata=metadata,
        cat_driver=mc.process.DatasetTransform,
        cat_path=mc.FILE_PATH_COMPILED(cat_orig.name),
        save_catalog=save_catalog,
        return_cat=True,
    )

    if return_cat:
        return cat


def setup(locs="mc_", override=False, boundaries=True):
    """Setup reference catalogs for models.

    Loops over catalogs that have been previously installed as data packages to intake that start with the string(s) in locs. The default is to read in the required GOODS model catalogs which are prefixed with `"mc_"`. Alternatively, one or more local catalog files can be input as strings or Paths.

    This function calls ``open_catalog`` which reads in previously-saved model boundary information (or calculates it if not available) and saves temporary catalog files for each model (called "compiled"), then this function links those together into the returned main catalog. For some models, reading in the original catalogs applies a "today" and/or "yesterday" date Intake user parameter that supplies two example model files that can be used for examining the model output for the example times. Those are rerun each time this function is rerun, filling the parameters using the proper dates.

    Note that saved compiled catalog files will be saved under the name inside the catalog, not the name of the file if you input a catalog path.

    Parameters
    ----------
    locs : str, Path, list
        This can be:

        * a string or Path describing where a Catalog file is located
        * a string of the prefix for selecting catalogs from the default intake catalog, ``intake.cat``. It is expected to be of the form "PREFIX_CATALOGNAME" with an underscore at the end followed by the catalog name, and there could be many catalogs with that `"PREFIX_"` set up.
        * a list of a combination of the previous options.

    override : boolean, optional
        Use `override=True` to compile the catalog files together regardless of freshness.
    boundaries : boolean, optional
        If True, find previously-saved or calculate domain boundary of model. Intended for testing.

    Returns
    -------
    Intake catalog
        Nested Intake catalog with a catalog for each input option. Each model in turn has one or more model_source available (e.g., "coops-forecast-agg", "coops-forecast-noagg").

    Examples
    --------

    Set up main catalog:

    >>> main_cat = mc.setup()

    Examine list of models available in catalog:

    >>> list(main_cat)

    Examine the model_sources for a specific model in the catalog:

    >>> list(main_cat['CBOFS'])

    Separate from ``model_catalogs`` you can check the default Intake catalog with:

    >>> list(intake.cat)
    """

    locs = mc.astype(locs, list)

    # arrange inputs into list of known Catalog instances and Paths to catalogs
    initial_cats = []
    for loc in locs:

        # initial_cats is a list of Catalogs in this case
        cats = [
            intake.cat[cat_name] for cat_name in list(intake.cat) if loc in cat_name
        ]

        # remove the prefix from the catalog name
        for cat in cats:
            cat.name = cat.name.lstrip(loc)

        # check for if loc is instead a path to a catalog
        if len(cats) == 0:
            # initial_cats is a list of one Path in this case
            # cats = [PurePath(loc)]
            # initial_cats is a list of one Catalog in this case
            cats = [intake.open_catalog(loc)]

        # now cats is a list of Catalog(s)
        initial_cats.extend(cats)

    cat_transform_locs = []
    for cat in list(initial_cats):

        # if isinstance(cat, PurePath):
        #     name = cat.stem
        # elif isinstance(cat, Catalog):
        name = cat.name

        # re-compile together catalog file if user wants to override possibly
        # existing file or if is not fresh
        if override or not mc.is_fresh(mc.FILE_PATH_COMPILED(name)):
            # override for open_catalog is about calculating boundaries
            open_catalog(
                cat,
                return_cat=False,
                save_catalog=True,
                boundaries=boundaries,
                save_boundaries=True,
                override=False,
            )
        cat_transform_locs.append(mc.FILE_PATH_COMPILED(name))

    # have to read these from disk in order to make them type
    # intake.catalog.local.YAMLFileCatalog
    # instead of intake.catalog.base.Catalog
    cats = [intake.open_catalog(loc) for loc in cat_transform_locs]

    # make master nested catalog
    main_cat = mc.make_catalog(
        cats,
        full_cat_name="MAIN-CATALOG",
        full_cat_description="Main catalog for models; a catalog of nested catalogs.",
        full_cat_metadata={},
        # full_cat_metadata={"source_catalog_dir": str(mc.CAT_PATH_ORIG)},
        cat_driver=intake.catalog.local.YAMLFileCatalog,
        cat_path=None,
        save_catalog=False,
    )

    return main_cat


def find_datetimes(source, find_start_datetime, find_end_datetime, override=False):
    """Find the start and/or end datetimes for source.

    For sources with static urlpaths, this opens the Dataset and checks the first time for `start_datetime` and the last time for `end_datetime`.

    Some NOAA OFS sources require aggregation: model_source names "coops-forecast-noagg" and "ncei-archive-noagg". For these, the available year and months of the thredd server subcatalogs are found with ``find_catrefs()``. `start_datetime` is found by further evaluating to make sure that files in the subcatalogs are both available on the page and that the days represented by model output files are consecutive (there are missing dates). `end_datetime` is found from the most recent subcatalog files since there aren't missing files and dates on the recent end of the time ranges.

    Uses ``cf-xarray`` to determine the time axis.

    Parameters
    ----------
    source : Intake source
        Model source for which to find start and/or end datetimes
    find_start_datetime : bool
        True to calculate start_datetime, otherwise returns None
    find_end_datetime : bool
        True to calculate end_datetime, otherwise returns None
    override : boolean, optional
        Use `override=True` to find catrefs regardless of freshness. This is passed in from
        ``find_availability()`` so has the same value as input there.

    Returns
    -------
    tuple
        Contains 'start_datetime' and 'end_datetime' where each are strings or can be None if they didn't need to be found.
    """

    # For any model/model_source pairs with static links or known file address,
    # which is all non-OFS models and OFS models that are already aggregated
    if "catloc" not in source.metadata:

        # flag to process.py for `to_dask()` call to skip detailed processing since just a quick check. Doesn't save ds to source._ds or cache.
        source._kwargs["skip_dask_processing"] = True
        ds = source.to_dask()
        # use one T in case there are more than one
        tkey = ds.cf.axes["T"][0]
        start_datetime = str(ds[tkey].values[0]) if find_start_datetime else None
        end_datetime = str(ds[tkey].values[-1]) if find_end_datetime else None
        ds.close()

    # for when we need to aggregate which is for model_source: ncei-archive-noagg and coops-forecast-noagg
    else:

        if "filetype" not in source.cat.metadata:
            raise KeyError(
                "If your model requires aggregation, it also requires `filetype` in the catalog-level metadata."
            )
        else:
            filetype = source.cat.metadata["filetype"]

        if not override and mc.is_fresh(
            mc.FILE_PATH_CATREFS(source.cat.name, source.name), source
        ):
            with open(
                mc.FILE_PATH_CATREFS(source.cat.name, source.name), "r"
            ) as stream:
                catrefs = yaml.safe_load(stream)["catrefs"]
        else:
            catrefs = mc.find_catrefs(source.metadata["catloc"])
            catrefs = sorted(catrefs)  # earliest first, most recent last
            with open(
                mc.FILE_PATH_CATREFS(source.cat.name, source.name), "w"
            ) as outfile:
                yaml.dump({"catrefs": catrefs}, outfile, default_flow_style=False)

        if find_start_datetime:
            # Getting start date #
            # first make sure the dates actually have model files available
            filelocs = []
            i = 0
            while len(filelocs) == 0:
                # print(catrefs[i])
                filelocs = sorted(
                    mc.find_filelocs(
                        catrefs[i], source.metadata["catloc"], filetype=filetype
                    )
                )
                i += 1

            # second make sure we only count when dates are consecutive, since servers tend to have some
            # spotty model output at the earliest dates get dates from file names
            df = mc.filedates2df(filelocs)

            # which differences in consecutive dates are over 1 day
            ddf = pd.Series(df.index).diff() > pd.Timedelta("1 day")
            if ddf.any():
                # first date after last jump in dates is desired start day
                start_datetime = str(df.index.where(ddf).dropna()[-1])

            # all dates were fine, so just use earliest fileloc
            else:
                start_datetime = str(df.index[0])
        else:
            start_datetime = None

        if find_end_datetime:
            # Getting end date #
            filelocs = mc.find_filelocs(
                catrefs[-1], source.metadata["catloc"], filetype=filetype
            )
            df = mc.filedates2df(filelocs)
            end_datetime = str(df.index[-1])
        else:
            end_datetime = None

    # save start/end to files
    if find_start_datetime:
        with open(mc.FILE_PATH_START(source.cat.name, source.name), "w") as outfile:
            yaml.dump(
                {"start_datetime": start_datetime}, outfile, default_flow_style=False
            )
    if find_end_datetime:
        with open(mc.FILE_PATH_END(source.cat.name, source.name), "w") as outfile:
            yaml.dump({"end_datetime": end_datetime}, outfile, default_flow_style=False)

    return start_datetime, end_datetime


def find_availability_source(source, override=False):
    """Find availabililty for source specifically.

    This function is called by `find_availability()` for each source. If ``source.status`` is False, input source is returned with None for `start_datetime` and `end_datetime`.

    Parameters
    ----------
    source : Intake source
        Source for which to find availability.

    Returns
    -------
    Intake source
        `start_datetime` and `end_datetime` are added to metadata of source.
    """

    # if server is not working, return input source with None for new metadata
    if not source.status:
        # raise RuntimeError(f"Server for source {source.cat.name}, {source.name}, is not working. Urlpath checked was {mc.astype(source.urlpath, list)[0]}.")
        warnings.warn(
            f"Server for source {source.cat.name}, {source.name}, is not working. Urlpath checked was {mc.astype(source.urlpath, list)[0]}.",
            RuntimeWarning,
        )
        start_datetime, end_datetime = None, None

    else:

        # check if start and end datetime files already exist and are new enough to use
        # check if already know the time and not stale
        # file times are given in UTC
        # If files are not stale, read in info from there
        if not override and mc.is_fresh(
            mc.FILE_PATH_START(source.cat.name, source.name), source
        ):
            with open(mc.FILE_PATH_START(source.cat.name, source.name), "r") as stream:
                start_datetime = yaml.safe_load(stream)["start_datetime"]
            find_start_datetime = False
        else:
            find_start_datetime = True  # need to still find the start_datetime

        if not override and mc.is_fresh(
            mc.FILE_PATH_END(source.cat.name, source.name), source
        ):
            with open(mc.FILE_PATH_END(source.cat.name, source.name), "r") as stream:
                end_datetime = yaml.safe_load(stream)["end_datetime"]
            find_end_datetime = False
        else:
            find_end_datetime = True  # need to still find the end_datetime

        # start and end temp could be None, depending on values of
        # find_start_datetime, find_end_datetime
        if find_start_datetime or find_end_datetime:
            start_temp, end_temp = find_datetimes(
                source, find_start_datetime, find_end_datetime, override=override
            )

        start_datetime = start_temp if find_start_datetime else start_datetime
        end_datetime = end_temp if find_end_datetime else end_datetime

    source.metadata["start_datetime"] = start_datetime
    source.metadata["end_datetime"] = end_datetime

    return source


def find_availability(cat_or_source, model_source=None, override=False, verbose=False):
    """Find availability for Catalog or Source.

    The code will check for previously-calculated availability. If found, the "freshness" of the information is checked as compared with ``mc.FRESH`` parameters specified in ``__init__``.

    Start and end datetimes are allowed to be calculated separately to save time.

    Note that for unaggregated models with forecasts (currently just model_source "coops-forecast-noagg"), this checks availability for the latest forecast, which goes forward in time from today. It is not possible to use this function to check for the case of a forecast forward in time from a past day.

    Parameters
    ----------
    cat_or_source : Intake catalog or source
        Catalog containing model_source sources for which to find availability, or single Source for which to find availability.
    model_source : str, list of strings, optional
        Specified model_source(s) for which to find the availability for a catalog. If unspecified and cat_or_source is a Catalog, loop over all sources in catalog and find availability for all.
    override : boolean, optional
        Use `override=True` to find availability regardless of freshness.
    verbose : boolean, optional
        If True, `start_datetime` and `end_datetime` found for each Source will be printed.

    Returns
    -------
    Intake catalog or source
        If a Catalog was input, a Catalog will be returned; if a Source was input, a Source will be returned. For the single input Source or all Sources in the input Catalog, `start_datetime` and `end_datetime` are added to metadata.

    Examples
    --------

    Set up source catalog, then find availability for all sources of CIOFS model:

    >>> main_cat = mc.setup()
    >>> cat = mc.find_availability(main_cat['CIOFS'], model_source=['coops-forecast-agg', 'coops-forecast-noagg'])

    Find availability for only model_source "coops-forecast-noagg" of CBOFS model, and print it:

    >>> source = mc.find_availability(main_cat['CBOFS']['coops-forecast-noagg'], verbose=True)
    coops-forecast-noagg: 2022-08-22 13:00:00 to  2022-09-25 12:00:00
    """

    # Check in case user input main_catalog which is not correct
    # if both the input obj and the items one layer down, contained in obj, are both catalogs, then
    # a nested catalog was input which is not correct
    if isinstance(cat_or_source, Catalog) and isinstance(
        cat_or_source[list(cat_or_source)[0]], Catalog
    ):
        raise ValueError(
            "A nested catalog was input, but should be either a catalog that contains sources instead of catalogs, or a source. For example, try `main_cat['CBOFS']` or `main_cat['CBOFS']['coops-forecast-agg']`."
        )

    # if Catalog was input
    if isinstance(cat_or_source, Catalog):
        # if no model_source input, loop through all
        model_sources = (
            list(cat_or_source)
            if model_source is None
            else mc.astype(model_source, list)
        )
        sources = []
        for model_source in model_sources:
            try:
                source = find_availability_source(
                    source=cat_or_source[model_source], override=override
                )
            except requests.exceptions.HTTPError:
                print(f"Found a server error with {model_source}")
                source = cat_or_source[model_source]
            # source = find_availability_source(
            #     source=cat_or_source[model_source], override=override
            # )
            sources.append(source)
            if verbose:
                print(
                    f"{source.name}: {source.metadata['start_datetime']} to {source.metadata['end_datetime']}"
                )

        # Make new catalog to remember the new metadata
        new_user_cat = mc.make_catalog(
            sources,
            full_cat_name=cat_or_source.name,
            full_cat_description=cat_or_source.description,
            full_cat_metadata=cat_or_source.metadata,
            cat_driver=[source._entry._driver for source in list(sources)],
            cat_path=None,
            save_catalog=False,
        )

        return new_user_cat

    # if Source input
    elif isinstance(cat_or_source, (OpenDapSource, DatasetTransform)):

        # doesn't make sense to input a model_source and source
        if model_source is not None:
            raise ValueError("A source was input, so `model_source` should be None.")

        source = find_availability_source(source=cat_or_source, override=override)

        if verbose:
            print(
                f"{source.name}: {source.metadata['start_datetime']} to  {source.metadata['end_datetime']}"
            )

        # Make new catalog to remember the new metadata, then extract source immediately
        out_source = mc.make_catalog(
            [source],
            full_cat_name=cat_or_source.cat.name,
            full_cat_description=cat_or_source.cat.description,
            full_cat_metadata=cat_or_source.cat.metadata,
            cat_driver=cat_or_source._entry._driver,
            cat_path=None,
            save_catalog=False,
        )[source.name]

        return out_source

    else:
        raise ValueError(
            f"Either an Intake catalog or Intake source should be input, but found object of type {type(cat_or_source)}."
        )


def transform_source(source_orig):
    """Set up transform of original catalog source

    Parameters
    ----------
    source_orig : Intake source
        Original source, which will be transformed

    Returns
    -------
    Intake source
        `source_transform`, the transformed version of source_orig. This source will point at the source of `source_orig` as the target.
    """

    # open the skeleton transform cat entry and then alter
    # a few things so can use it with source_orig
    source_transform = intake.open_catalog(mc.CAT_PATH_TRANSFORM)["name"]

    # Update name and description for transformed source
    source_transform.name = source_orig.name
    source_transform.description = (
        f"Catalog entry for transform of dataset {source_orig.name}"
    )

    # make path to source_orig the target
    source_transform._captured_init_kwargs["targets"] = [
        f"{source_orig.cat.path}:{source_orig.name}"
    ]
    source_transform.targets = [f"{source_orig.cat.path}:{source_orig.name}"]

    # add metadata from source_orig
    source_transform.metadata.update(source_orig.metadata)

    return source_transform


def select_date_range(
    cat_or_source,
    start_date,
    end_date=None,
    model_source=None,
    use_forecast_files=None,
    override=False,
):
    """For NOAA OFS unaggregated models: Update `urlpath` locations in `Source`.

    For other models, set up so that `start_date` and `end_date` are used to filter resulting Dataset in time. For all models, save `start_date` and `end_date` in the `Source` metadata.

    NOAA OFS model sources that require aggregation (currently for model_sources "coops-forecast-noagg" and "ncei-archive-noagg") need to have the specific file paths found for each file that will be read in. This function does that, based on the desired date range, and returns a `Source` with file locations in the `urlpath`. This function can also be used with any model that does not require this (because the model paths are either static or deterministic) but in those cases it does not need to be used; they will have the start and end dates applied to filter the resulting model output after ``to_dask()`` is called.

    Parameters
    ----------
    cat_or_source : Intake catalog or source
        Catalog containing model_source sources, or single Source.
    start_date: datetime-interpretable str or pd.Timestamp
        Date (and possibly time) of start to desired model date range. If input date does not include a time, times will be included from the start of the day. If a time is input in `start_date`, it is used to narrow the time range of the results.
    end_date: datetime-interpretable str, pd.Timestamp, or None; optional
        Date (and possibly time) of start to desired model date range. If input date does not include a time, times will be included from the start of the day. If a time is input in `start_date`, it is used to narrow the time range of the results. end_date can be None which indicates the user wants all available model output after start_date; this optional is not available for unaggregated historical NOAA OFS models which do not contain forecast files (i.e., model_source "ncei-archive-noagg").

        There are several use cases to specify:

        * if `start_date == end_date`, the full day of model output from the date is selected. If the date specified is today and all times for today are not yet available, output from forecast files will be used to fill out the day after the nowcast files end.
        * If `end_date is None`, all available model output will be retrieved starting at start_date. This option doesn't work for archival unaggregated NOAA OFS models currently.
        * If `end_date` is in the future, `use_forecast_files` is set to True and the forecast is read in, but stopped at `end_date`.
        * User can set `use_forecast_files=True` with an end_date in the past to get old forecast model results for end_date for unaggregated NOAA OFS models. This case is probably not well-used and is not regularly tested. The results from using this combination of inputs does not align with the results of ``mc.find_availability()`` since the forecast is not the latest.
    model_source: str, optional
        Which model_source to use. If ``mc.find_availability()`` has been run, the code will determine which model_source in the Catalog to use based on `start_date` and `end_date`. Otherwise a single model_source can be provided, or ``find_availability()`` will be run if needed. An exception is if there is only one model_source available for cat, that one will be used without specifying it.
    use_forecast_files : bool or None, optional
        This parameter is typically set by the code and is not used by the user. However, in one use case the user can input `use_forecast_files=True`: when they want to read in a forecast from the past for a NOAA OFS model. Otherwise do not use this parameter directly.
    override : boolean, optional
        Use `override=True` to find catrefs regardless of freshness.

    Returns
    -------
    Intake Source
        Intake `Source` associated with the catalog entry which now contains `source.metadata['start_date']` and `source.metadata['end_date']`. The values of `source.metadata['start/end_date']` will not necessarily be the same as the input `start_date` and `end_date`, but may be changed to return the desired output time range. For unaggregated NOAA OFS models, the returned Source will have updated `source.urlpath` to reflect the newly-found file paths of the selected date range.

    Examples
    --------

    Find model `'LMHOFS'` urlpaths for all of today through all available forecast, directly from source catalog without first searching for availability with ``mc.find_availability()``:

    >>> main_cat = mc.setup()
    >>> today = pd.Timestamp.today()
    >>> source = mc.select_date_range(main_cat["LMHOFS"]["coops-forecast-noagg"], start_date=today, end_date=None)

    Find urlpaths with ``select_date_range`` and have it run ``find_availability()``:

    >>> source = mc.select_date_range(main_cat['LMHOFS'], start_date=today, end_date=today)

    """

    # Check in case user input main_catalog which is not correct
    # if both the input obj and the items one layer down, contained in obj, are both catalogs, then
    # a nested catalog was input which is not correct
    if isinstance(cat_or_source, Catalog) and isinstance(
        cat_or_source[list(cat_or_source)[0]], Catalog
    ):
        raise ValueError(
            "A nested catalog was input, but should be either a catalog that contains sources instead of catalogs, or a source. For example, try `main_cat['CBOFS']` or `main_cat['CBOFS']['coops-forecast-agg']`."
        )

    # save these to determine if user input dates with times or not
    start_date_input, end_date_input = start_date, end_date

    # make sure start_date and end_date are both Timestamps
    start_date = mc.astype(start_date, pd.Timestamp)
    end_date = None if end_date is None else mc.astype(end_date, pd.Timestamp)

    today = pd.Timestamp.today()

    if end_date is None and model_source == "ncei-archive-noagg":
        raise KeyError(
            "model_source 'ncei-archive-noagg' does not have forecast files, so `end_date` should be a datetime representation."
        )

    if end_date is not None and end_date.date() < start_date.date():
        raise KeyError(
            f"`start_date` is {start_date} but needs to be earlier than `end_date` ({end_date})."
        )

    # upon input, use_forecast_files should only be True if end_date is not None and end_date < today
    # later, however, it will be set to True in several circumstances
    if use_forecast_files and (end_date is None or end_date.date() >= today.date()):
        warnings.warn(
            f"User should only input `use_forecast_files=True` to access a forecast from the past, however `end_date={end_date}`.",
            UserWarning,
        )

    # set start/end_date_sel, but end_date_sel is overwritten in one unusual case subsequently
    if start_date == end_date is not None:
        start_date_sel = str(start_date.date())
        end_date_sel = str(end_date.date())

    else:
        # If end_date_input contains the default input time options from dateutil, assume time not input
        # in which case use date only to retrieve the whole day of output when selecting
        # or end_date_input exactly matches start_date_input
        if (
            end_date_input is not None
            and parse(mc.astype(end_date_input, str), default=DEFAULT).strftime(
                "%H%M%S"
            )
            == "222222"
        ):
            # user didn't specify time or end_date is None
            end_date_sel = str(end_date.date())
        else:  # user specified time
            end_date_sel = str(end_date) if end_date is not None else None

        # same but for start_date at the beginning of the day
        if (
            parse(mc.astype(start_date_input, str), default=DEFAULT).strftime("%H%M%S")
            == "222222"
        ):
            start_date_sel = str(start_date.date())
        else:
            start_date_sel = str(start_date) if start_date is not None else None

    # set end_date_loop
    # if end_date is None, set use_forecast_files to True and end_date_loop to today
    # if end_date is after today, code set use_forecast_files to True and end_date_loop to today
    if (end_date is None) or (end_date.date() > today.date()):
        end_date_loop = today
        use_forecast_files = True
    # is end_date before today, set end_date_loop forward 1 day to get last few time steps of the day
    elif end_date.date() < today.date():
        # In only this case, since user inputs use_forecast_files as True,  end_date_loop should be end_date and should retrieve forecast from yesterday
        if use_forecast_files:
            end_date_loop = end_date
            # This is out of place! Overwrite end_date_sel with None
            end_date_sel = None
        else:
            end_date_loop = end_date + pd.Timedelta("1 day")
    # if end_date is today, set end_date_loop to end_date and use_forecast_files True to get the last few time steps of today
    elif end_date.date() == today.date():
        end_date_loop = end_date
        use_forecast_files = True

    # if Catalog was input, determine which model_source to use to get to a source
    if isinstance(cat_or_source, Catalog):

        cat = cat_or_source

        # if there is only one model_source, use it
        if model_source is None and len(list(cat)) == 1:
            model_source = list(cat)[0]

        # is model_source was input, make sure it is in the Catalog
        if model_source is not None and model_source not in list(cat):
            raise KeyError(
                f"User input model_source {model_source} but it is not a source in catalog {cat.name}."  # noqa: E501
            )

        elif model_source is None and any(
            [
                "start_datetime" not in cat[model_source].metadata
                or "end_datetime" not in cat[model_source].metadata
                for model_source in list(cat)
            ]
        ):
            warnings.warn(
                "`find_availability()` has not been run for this model — running now..."
            )
            cat = mc.find_availability(cat, verbose=True)
            # raise KeyError(
            #     "Run `mc.find_availability()` for this model before running this command. Otherwise input model_source that contains desired date range."  # noqa: E501
            # )

        # which source to use from catalog for desired date range
        if model_source is None:
            user_range = DateTimeRange(start_date, end_date)

            for model_source in list(cat):
                model_source_range = DateTimeRange(
                    cat[model_source].metadata["start_datetime"],
                    cat[model_source].metadata["end_datetime"],
                )
                try:  # use this model_source if it is in the date range
                    if user_range in model_source_range:
                        break
                except TypeError:
                    continue
            else:
                raise ValueError(
                    "date range does not fully fit into any model model_sources"
                )

        if use_forecast_files and model_source == "ncei-archive-noagg":
            raise KeyError(
                "model_source 'ncei-archive-noagg' does not have forecast files, so `use_forecast_files` should be False."
            )

        source = cat[model_source]

    # if Source input
    elif isinstance(cat_or_source, (OpenDapSource, DatasetTransform)):

        source = cat_or_source
        model_source = source.name

    # if server is not working, return input source with None for new metadata
    if not source.status:
        # raise RuntimeError(f"Server for source {source.cat.name}, {source.name}, is not working. Urlpath checked was {mc.astype(source.urlpath, list)[0]}.")
        warnings.warn(
            f"Server for source {source.cat.name}, {source.name}, is not working. Urlpath checked was {mc.astype(source.urlpath, list)[0]}.",
            RuntimeWarning,
        )
        return source

    # catch the models that require aggregation
    if "catloc" in source.metadata:
        pattern = source.metadata["pattern"] if "pattern" in source.metadata else None

        # check catalog
        if not mc.status(source.metadata["catloc"], suffix=""):
            raise RuntimeError(
                f"Catalog {source.metadata['catloc']} for source {source.cat.name}, {source.name}, is not working."
            )
            # warnings.warn(
            #     f"Catalog {source.metadata['catloc']} for source {source.cat.name}, {source.name}, is not working.",
            #     RuntimeWarning,
            # )
            # return None

        if not override and mc.is_fresh(
            mc.FILE_PATH_CATREFS(source.cat.name, source.name), source
        ):
            with open(
                mc.FILE_PATH_CATREFS(source.cat.name, source.name), "r"
            ) as stream:
                catrefs = yaml.safe_load(stream)["catrefs"]
        else:
            try:
                catrefs = mc.find_catrefs(source.metadata["catloc"])
            except Exception as e:
                print(e)
            catrefs = sorted(catrefs)  # earliest first, most recent last
            with open(
                mc.FILE_PATH_CATREFS(source.cat.name, source.name), "w"
            ) as outfile:
                yaml.dump({"catrefs": catrefs}, outfile, default_flow_style=False)

        # loop over dates
        filelocs_urlpath = []
        for date in pd.date_range(
            start=start_date.normalize(), end=end_date_loop, freq="1D"
        ):
            is_forecast = (
                True
                if date.date() == end_date_loop.date() and use_forecast_files
                else False
            )

            fname = mc.FILE_PATH_AGG_FILE_LOCS(
                source.cat.name, model_source, date, is_forecast
            )

            if not override and mc.is_fresh(fname, source):
                with open(fname, "r") as stream:
                    agg_filelocs = yaml.safe_load(stream)["agg_filelocs"]

            else:
                # translate date to catrefs to select which catref to use
                if len(catrefs[0]) == 3:
                    cat_ref_to_match = (
                        date.strftime("%Y"),
                        date.strftime("%m"),
                        date.strftime("%d"),
                    )
                elif len(catrefs[0]) == 2:
                    cat_ref_to_match = (date.strftime("%Y"), date.strftime("%m"))

                if cat_ref_to_match in catrefs:
                    ind = catrefs.index(cat_ref_to_match)
                else:
                    warnings.warn(
                        f"Probably the time range requested is not available for model {source.cat.name}, model_source {source.name}. Returning source now.",
                        RuntimeWarning,
                    )
                    return source

                filelocs = mc.find_filelocs(
                    catrefs[ind],
                    source.metadata["catloc"],
                    source.cat.metadata["filetype"],
                )

                agg_filelocs = mc.agg_for_date(
                    date,
                    filelocs,
                    source.cat.metadata["filetype"],
                    is_forecast,
                    pattern,
                )

                with open(fname, "w") as outfile:
                    yaml.dump(
                        {"agg_filelocs": agg_filelocs},
                        outfile,
                        default_flow_style=False,
                    )

            filelocs_urlpath.extend(agg_filelocs)

        # df is sorted and deduplicated
        df = mc.filedates2df(filelocs_urlpath)

        # Narrow the files used to the actual requested datetime range
        files_to_use = df[start_date_sel:end_date_sel]

        # get only unique files and change to list
        files_to_use = list(pd.unique(files_to_use["filenames"]))

        # This is how we input the newly found urlpaths in so they will be used
        # in the processing of the dataset, and overwrite the old urlpath
        source._captured_init_kwargs["transform_kwargs"]["urlpath"] = files_to_use

        # Then run the transform for urlpath to pass that info on
        source.update_urlpath()

    # Pass start and end dates to the transform so they can be implemented
    # there for static and deterministic model files (includes RTOFS) as well
    # as the OFS aggregated models.
    source._captured_init_kwargs["transform_kwargs"]["start_date"] = start_date_sel
    source._captured_init_kwargs["transform_kwargs"]["end_date"] = end_date_sel

    # store info in source_orig
    metadata = {
        "model_source": model_source,
        "start_date": start_date_sel,
        "end_date": end_date_sel,
    }
    source.metadata.update(metadata)
    # Add original overall model catalog metadata to this next version
    source.metadata.update(source.cat.metadata)

    return source



"""
Set up for using package.
"""

import importlib

from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

import pandas as pd

from appdirs import AppDirs

from .model_catalogs import (  # noqa
    find_availability,
    make_catalog,
    open_catalog,
    select_date_range,
    setup,
    transform_source,
)
from .utils import (  # noqa
    agg_for_date,
    astype,
    calculate_boundaries,
    file2dt,
    filedates2df,
    find_bbox,
    find_catrefs,
    find_filelocs,
    get_fresh_parameter,
    is_fresh,
    status,
)


try:
    __version__ = version("model_catalogs")
except PackageNotFoundError:
    # package is not installed
    __version__ = "unknown"

# this forces single threading which avoids an issue described here:
# https://github.com/pydata/xarray/issues/7079
# https://github.com/Unidata/netcdf4-python/issues/1192
import dask


dask.config.set(scheduler="single-threaded")

# set up known locations for catalogs.
# this is where the original model catalog files and previously-calculated
# model boundaries can be found, which are hard-wired in the repo
# version change for this behavior
try:  # >= Python 3.9
    CAT_PATH = importlib.resources.files("model_catalogs") / "support_files"
except AttributeError:  # < Python3.9
    with importlib.resources.path("model_catalogs", "support_files") as pth:
        CAT_PATH = pth
# CAT_PATH_ORIG = CAT_PATH / "orig"
# CAT_PATH_BOUNDARIES = CAT_PATH / "boundaries"
CAT_PATH_TRANSFORM = CAT_PATH / "transform.yaml"

# test files
try:  # >= Python 3.9
    TEST_PATH = importlib.resources.files("model_catalogs") / "tests"
except AttributeError:  # < Python3.9
    with importlib.resources.path("model_catalogs", "tests") as pth:
        TEST_PATH = pth
TEST_PATH_FILES = TEST_PATH / "test_files"

# set up cache directories for package to use
# user application cache directory, appropriate to each OS
dirs = AppDirs("model_catalogs", "NOAA-ORR-ERD")
cache_dir = Path(dirs.user_cache_dir)


# This is where files are saved to refer back to for saving time
CACHE_PATH = cache_dir
CACHE_PATH_COMPILED = CACHE_PATH / "compiled"
CACHE_PATH_AVAILABILITY = CACHE_PATH / "availability"
CACHE_PATH_FILE_LOCS = CACHE_PATH / "file_locs"

# whenever a package of catalogs is installed, have it copy any available boundaries to here
CAT_PATH_BOUNDARIES = CACHE_PATH / "boundaries"

# make directories
CACHE_PATH_COMPILED.mkdir(parents=True, exist_ok=True)
CACHE_PATH_AVAILABILITY.mkdir(parents=True, exist_ok=True)
CACHE_PATH_FILE_LOCS.mkdir(parents=True, exist_ok=True)
CAT_PATH_BOUNDARIES.mkdir(parents=True, exist_ok=True)


def TEST_PATH_FILE(model, model_source):
    """Return file path to test file."""
    model = model.lower().replace("-", "_")
    return (TEST_PATH_FILES / f"{model}_{model_source}").with_suffix(".nc")


def FILE_PATH_COMPILED(model):
    """Return filename for model boundaries information."""
    return (CACHE_PATH_COMPILED / model).with_suffix(".yaml")


# availability file names
def FILE_PATH_START(model, model_source):
    """Return filename for model/model_source start time."""
    return CACHE_PATH_AVAILABILITY / f"{model}_{model_source}_start_datetime.yaml"


def FILE_PATH_END(model, model_source):
    """Return filename for model/model_source end time."""
    return CACHE_PATH_AVAILABILITY / f"{model}_{model_source}_end_datetime.yaml"


def FILE_PATH_BOUNDARIES(model):
    """Return filename for model boundaries information."""
    return (CAT_PATH_BOUNDARIES / model).with_suffix(".yaml")


def FILE_PATH_CATREFS(model, model_source):
    """Return filename for model/model_source start time."""
    return CACHE_PATH_AVAILABILITY / f"{model}_{model_source}_catrefs.yaml"


def FILE_PATH_AGG_FILE_LOCS(model, model_source, date, is_fore):
    """Return filename for aggregated file locations.

    Date included to day."""
    date = astype(date, pd.Timestamp)
    name = f"{model}_{model_source}_{date.isoformat()[:10]}_is-forecast_{is_fore}.yaml"
    return CACHE_PATH_FILE_LOCS / name


# Fresh parameters: how long until model output avialability will be refreshed
# for `find_availabililty()` if requested
FRESH = {
    "start": "1 day",
    "end": "4 hours",
    "catrefs": "6 hours",
    "file_locs": "4 hours",
    "compiled": "6 hours",  # want to be on the same calendar day as when they were compiled; this approximates that.  # noqa: E501
}



"""
Make sure catalogs work correctly. If this doesn't run once, try again since the servers can be finnicky.
"""

import warnings

from unittest import mock

import intake
import numpy as np
import pandas as pd
import pytest
import xarray as xr

from intake.catalog import Catalog
from intake_xarray.opendap import OpenDapSource
from pandas import Timestamp

import model_catalogs as mc

from model_catalogs import process
from model_catalogs.process import DatasetTransform


def test_setup():
    """Make sure main catalog is created correctly."""

    main_cat = mc.setup(override=True, boundaries=False)

    # check that all compiled catalog files exist
    should_exist = [
        cat.lstrip("mc_") for cat in list(intake.cat) if cat.startswith("mc_")
    ]
    assert sorted(list(main_cat)) == sorted(should_exist)

    # Check that model_sources are correct for one test case
    assert sorted(list(main_cat["CBOFS"])) == [
        "coops-forecast-agg",
        "coops-forecast-noagg",
        "ncei-archive-noagg",
    ]
    # assert main_cat["CBOFS"].metadata["geospatial_bounds"]


# @pytest.mark.slow
def test_find_availability():
    """Test find_availability.

    Using warnings instead of exception since sometimes servers aren't reliable.
    """

    # test models with fast static links and that require aggregations
    test_models = {
        "GOFS": "hycom-forecast-agg",
        "DBOFS": "coops-forecast-noagg",
        "SFBOFS": "ncei-archive-noagg",
    }

    main_cat = mc.setup(boundaries=False)
    for model, model_source in test_models.items():
        cat = mc.find_availability(
            main_cat[model], model_source=model_source, override=True
        )
        if "start_datetime" not in cat[model_source].metadata:
            warnings.warn(
                f"Running model {model} with model_source {model_source} in `find_availability()` did not result in `start_datetime` in the catalog metadata.",  # noqa: E501
                RuntimeWarning,
            )
        fname = mc.FILE_PATH_START(model, model_source)
        if not mc.is_fresh(fname, cat[model_source]):
            warnings.warn(f"Filename {fname} is not found as fresh.", RuntimeWarning)

        # make sure catalog output since catalog was input
        assert isinstance(cat, Catalog)

        # also compare with requesting source directly
        source = mc.find_availability(main_cat[model][model_source])

        # make sure source output since source was input
        assert isinstance(source, (OpenDapSource, DatasetTransform))

        #         assert (
        #             cat[model_source].metadata["start_datetime"]
        #             == source.metadata["start_datetime"]
        #         )
        #         assert (
        #             cat[model_source].metadata["end_datetime"]
        #             == source.metadata["end_datetime"]
        #         )

        # test that if server status is False, start_datetime, end_datetime are None
        in_source = main_cat[model][model_source]
        in_source._status = False
        with pytest.warns(RuntimeWarning):
            out_source = mc.find_availability(in_source)
        assert out_source.metadata["start_datetime"] is None
        assert out_source.metadata["end_datetime"] is None


# @pytest.mark.slow
def test_boundaries():
    """Test one faster model and make sure variables come through."""

    model = "mc_GOFS"
    cat = intake.cat[model]

    # Calculate
    boundaries = mc.calculate_boundaries(
        cats=cat, save_files=False, return_boundaries=True
    )

    assert "bbox" in boundaries[model]
    assert "wkt" in boundaries[model]


# @pytest.mark.slow
# this test continues to be too brittle with CO-OPS servers breaking constantly but in different ways each time
# def test_select_date_range():
#     """Test functionality in `select_date_range()`.

#     Should filter date range for static and nonstatic links. Should return all model output for today
#     and tomorrow.
#     """

#     test_models = {"GOFS": "hycom-forecast-agg", "CIOFS": "coops-forecast-noagg"}

#     today = pd.Timestamp.today() - pd.Timedelta("1 day")
#     # today = pd.Timestamp.today(tz="UTC") - pd.Timedelta("1 day")
#     tom = today + pd.Timedelta("1 day")

#     main_cat = mc.setup()
#     for model, model_source in test_models.items():
#         try:
#             source = mc.select_date_range(
#                 main_cat[model][model_source], today.date(), tom.date(), override=True
#             )
#         except RuntimeError as e:
#             warnings.warn(
#                 f"Source {model}, {model_source} had runtime issue with exception {e}.",  # noqa: E501
#                 RuntimeWarning,
#             )


#         # if source.status:
#         try:
#             # import pdb; pdb.set_trace()
#             ds = source.to_dask()

#             assert ds.cf["T"][0] == today.normalize()

#             # calculate delta times
#             dts = ds.cf["T"].diff(dim=ds.cf.axes["T"][0])

#             # make sure all dts are the same (some are 42 seconds off for some reason but that is ok)
#             assert all(
#                 [
#                     pd.Timedelta(f"{float(dt)} {np.datetime_data(dt)[0]}")
#                     < pd.Timedelta("1 minute")
#                     for dt in dts - dts[0]
#                 ]
#             )

#             end_of_day = tom.normalize() + pd.Timedelta("1 day")
#             assert bool(
#                 end_of_day
#                 - pd.Timedelta(f"{float(dts[0])} {np.datetime_data(dts[0])[0]}")
#                 <= ds.cf["T"][-1]
#                 < end_of_day
#             )

#         except OSError as e:
#             warnings.warn(
#                 f"Some aource {model}, {model_source} file is unavailable.",  # noqa: E501
#                 RuntimeWarning,
#             )


#         # else:
#         #     warnings.warn(
#         #         f"Source {model}, {model_source} server status is False.",  # noqa: E501
#         #         RuntimeWarning,
#         #     )

#     # also make sure an incorrect requested datetime range returns a warning
#     # check this for static link models
#     test_models = {"GOFS": "hycom-forecast-agg", "CIOFS": "coops-forecast-agg"}

#     main_cat = mc.setup()
#     for model, model_source in test_models.items():
#         source = mc.select_date_range(
#             main_cat[model][model_source], "1980-1-1", "1980-1-2"
#         )

#         with pytest.raises(RuntimeError):
#             ds = source.to_dask()


# this test continues to be too brittle with CO-OPS servers breaking constantly but in different ways each time
# def test_select_date_range_dates():
#     """Detailed tests for resulting start/end datetimes.

#     Doesn't run `to_dask()`, and only checks unaggregated models.
#     All the tested models have hourly output.
#     """

#     yes = (
#         pd.Timestamp.today().normalize()
#         - pd.Timedelta("1 day")
#         + pd.Timedelta("6:00:00")
#     )
#     yes_date = str(yes.date())
#     yes_st = yes.normalize()
#     yes_end = yes_st + pd.Timedelta("23:00:00")
#     tod = pd.Timestamp.today().normalize() + pd.Timedelta("6:00:00")

#     test_models = {"CBOFS": "coops-forecast-noagg", "NYOFS": "coops-forecast-noagg"}
#     test_conditions = [
#         {"sday": yes, "eday": yes, "tst_known": yes_st, "tend_known": yes_end},
#         {
#             "sday": yes_date,
#             "eday": yes_date,
#             "tst_known": yes_st,
#             "tend_known": yes_end,
#         },
#         {"sday": yes, "eday": None, "tst_known": yes, "tend_known": None},
#         {"sday": tod, "eday": None, "tst_known": tod, "tend_known": None},
#     ]

#     main_cat = mc.setup()
#     for model, model_source in test_models.items():
#         for tc in test_conditions:
#             if tc["tend_known"] is None:
#                 cat = mc.find_availability(
#                     main_cat[model], model_source=model_source, override=True
#                 )
#             else:
#                 cat = main_cat[model]

#             try:
#                 source = mc.select_date_range(
#                     cat[model_source],
#                     start_date=tc["sday"],
#                     end_date=tc["eday"],
#                     override=True,
#                 )
#             except RuntimeError:
#                 print(f"server isn't working correctly right now for {cat.name}, {model_source}")

#             assert source.dates[0] == tc["tst_known"]
#             if tc["tend_known"] is None:
#                 assert source.dates[-1] == pd.Timestamp(source.metadata["end_datetime"])
#             else:
#                 assert source.dates[-1] == tc["tend_known"]

#             # check that dates are all consistent
#             ddf = pd.Series(source.dates).diff()
#             assert (ddf[1:] - ddf.median() < pd.Timedelta("1 min")).all()

#     # check archive, which stopped 10/22/22 and does not have forecast files
#     date = pd.Timestamp("2022-10-20T06:00")
#     date_date = str(date.date())
#     date_st = date.normalize()
#     date_end = date_st + pd.Timedelta("23:00:00")

#     test_models = {"SFBOFS": "ncei-archive-noagg"}
#     # check t-1_known of None with find_availability output
#     test_conditions = [
#         {
#             "sday": date,
#             "eday": date,
#             "tst_known": date_st,
#             "tend_known": date_end,
#         },
#         {
#             "sday": date_date,
#             "eday": date_date,
#             "tst_known": date_st,
#             "tend_known": date_end,
#         },
#     ]

#     main_cat = mc.setup()
#     for model, model_source in test_models.items():
#         for tc in test_conditions:
#             if tc["tend_known"] is None:
#                 cat = mc.find_availability(
#                     main_cat[model], model_source=model_source, override=True
#                 )
#             else:
#                 cat = main_cat[model]
#             source = mc.select_date_range(
#                 cat[model_source],
#                 start_date=tc["sday"],
#                 end_date=tc["eday"],
#                 override=True,
#             )

#             assert source.dates[0] == tc["tst_known"]
#             if tc["tend_known"] is None:
#                 assert source.dates[-1] == pd.Timestamp(source.metadata["end_datetime"])
#             else:
#                 assert source.dates[-1] == tc["tend_known"]

#             # check that dates are all consistent
#             ddf = pd.Series(source.dates).diff()
#             assert (ddf[1:] - ddf.median() < pd.Timedelta("1 min")).all()


# @pytest.mark.slow
def test_process():
    """Test that dataset is processed."""

    main_cat = mc.setup(boundaries=False)

    # if this dataset hasn't been processed, lon and lat won't be in coords
    assert "lon" in list(
        main_cat["LOOFS_FVCOM"]["coops-forecast-noagg"].to_dask().coords
    )


def check_source(source):
    """Check attributes of source for other tests."""

    try:
        if source.status:
            ds = source.to_dask()
        else:
            warnings.warn(
                f"Source {source.cat.name}, {source.name} server status is False.",
                RuntimeWarning,
            )
            return
    except Exception as e:
        warnings.warn(
            f"Source {source.cat.name}, {source.name} could not be read in by `xarray`, with uncaught exception: {e}."
        )
        return
    # except OSError:
    #     warnings.warn(
    #         f"Model {source.cat.name} with model_source {source.name} is not working right now.",
    #         RuntimeWarning,
    #     )
    #     return

    # check axis attributes have been assigned
    checks = [
        ds[var_name].attrs["axis"] == axis
        for axis, var_names in source.metadata["axis"].items()
        for var_name in mc.astype(var_names, list)
        if var_name in ds.dims
    ]
    assert all(checks)

    # check standard_names attributes have been assigned
    checks = [
        ds[var_name].attrs["standard_name"] == st_name
        for st_name, var_names in source.metadata["standard_names"].items()
        for var_name in mc.astype(var_names, list)
        if var_name in ds.data_vars
    ]
    assert all(checks)

    # check cf-xarray
    # AXIS Z won't be defined for unstructured 2D model output
    if (
        "SELFE" in source.target.description or "FVCOM" in source.target.description
    ) and "2DS" in source.cat.name:
        assert sorted(list(ds.cf.axes.keys())) == ["T", "X", "Y"]
    else:
        assert sorted(list(ds.cf.axes.keys())) == ["T", "X", "Y", "Z"]

    # the 2D cases are weird
    if "NGOFS2_2DS" in source.cat.name:
        assert sorted(list(ds.cf.coordinates.keys())) == [
            "latitude",
            "longitude",
            "time",
        ]
    else:
        assert sorted(list(ds.cf.coordinates.keys())) == [
            "latitude",
            "longitude",
            "time",
            "vertical",
        ]
    ds.close()


@pytest.mark.slow
def test_sources():
    """Test all known model sources.

    Fails gracefully. Does not require running `select_date_range()` because sources always have some
    known files included or a static link.
    """

    main_cat = mc.setup(boundaries=False)

    for model in list(main_cat):
        for model_source in list(main_cat[model]):
            source = main_cat[model][model_source]
            try:
                check_source(source)
            except AssertionError:
                warnings.warn(
                    f"Model {model} with model_source {model_source} does not have proper attributes.",
                    RuntimeWarning,
                )


def test_urlpath():
    """urlpath is exposed to transform from target now."""

    main_cat = mc.setup()

    # shows 2 default files
    assert len(main_cat["NGOFS2"]["coops-forecast-noagg"].urlpath) == 2


def test_file2dt():
    """Test a few sample files."""

    fname = "nos.cbofs.fields.n001.20220913.t00z.nc"
    date = Timestamp("20220912T19:00")
    assert mc.file2dt(fname) == date

    fname = "nos.cbofs.fields.f001.20220914.t12z.nc"
    date = Timestamp("20220914T13:00")
    assert mc.file2dt(fname) == date

    fname = "glofs.loofs.fields.nowcast.20220919.t00z.nc"
    date = [
        Timestamp("2022-09-18 19:00:00"),
        Timestamp("2022-09-18 20:00:00"),
        Timestamp("2022-09-18 21:00:00"),
        Timestamp("2022-09-18 22:00:00"),
        Timestamp("2022-09-18 23:00:00"),
        Timestamp("2022-09-19 00:00:00"),
    ]
    assert mc.file2dt(fname) == date

    fname = "nos.nyofs.fields.forecast.20220920.t17z.nc"
    date = [
        Timestamp("2022-09-20 18:00:00"),
        Timestamp("2022-09-20 19:00:00"),
        Timestamp("2022-09-20 20:00:00"),
        Timestamp("2022-09-20 21:00:00"),
        Timestamp("2022-09-20 22:00:00"),
        Timestamp("2022-09-20 23:00:00"),
        Timestamp("2022-09-21 00:00:00"),
        Timestamp("2022-09-21 01:00:00"),
        Timestamp("2022-09-21 02:00:00"),
        Timestamp("2022-09-21 03:00:00"),
        Timestamp("2022-09-21 04:00:00"),
        Timestamp("2022-09-21 05:00:00"),
        Timestamp("2022-09-21 06:00:00"),
        Timestamp("2022-09-21 07:00:00"),
        Timestamp("2022-09-21 08:00:00"),
        Timestamp("2022-09-21 09:00:00"),
        Timestamp("2022-09-21 10:00:00"),
        Timestamp("2022-09-21 11:00:00"),
        Timestamp("2022-09-21 12:00:00"),
        Timestamp("2022-09-21 13:00:00"),
        Timestamp("2022-09-21 14:00:00"),
        Timestamp("2022-09-21 15:00:00"),
        Timestamp("2022-09-21 16:00:00"),
        Timestamp("2022-09-21 17:00:00"),
        Timestamp("2022-09-21 18:00:00"),
        Timestamp("2022-09-21 19:00:00"),
        Timestamp("2022-09-21 20:00:00"),
        Timestamp("2022-09-21 21:00:00"),
        Timestamp("2022-09-21 22:00:00"),
        Timestamp("2022-09-21 23:00:00"),
        Timestamp("2022-09-22 00:00:00"),
        Timestamp("2022-09-22 01:00:00"),
        Timestamp("2022-09-22 02:00:00"),
        Timestamp("2022-09-22 03:00:00"),
        Timestamp("2022-09-22 04:00:00"),
        Timestamp("2022-09-22 05:00:00"),
        Timestamp("2022-09-22 06:00:00"),
        Timestamp("2022-09-22 07:00:00"),
        Timestamp("2022-09-22 08:00:00"),
        Timestamp("2022-09-22 09:00:00"),
        Timestamp("2022-09-22 10:00:00"),
        Timestamp("2022-09-22 11:00:00"),
        Timestamp("2022-09-22 12:00:00"),
        Timestamp("2022-09-22 13:00:00"),
        Timestamp("2022-09-22 14:00:00"),
        Timestamp("2022-09-22 15:00:00"),
        Timestamp("2022-09-22 16:00:00"),
        Timestamp("2022-09-22 17:00:00"),
        Timestamp("2022-09-22 18:00:00"),
        Timestamp("2022-09-22 19:00:00"),
        Timestamp("2022-09-22 20:00:00"),
        Timestamp("2022-09-22 21:00:00"),
        Timestamp("2022-09-22 22:00:00"),
        Timestamp("2022-09-22 23:00:00"),
    ]
    assert mc.file2dt(fname) == date


def test_filedates2df():
    """Checking sorting and deduplicating indices."""

    fnames = [
        "nos.creofs.fields.f000.20220920.t15z.nc",
        "nos.creofs.fields.f001.20220920.t15z.nc",
        "nos.creofs.fields.f002.20220920.t15z.nc",
        "nos.creofs.fields.f003.20220920.t15z.nc",
        "nos.creofs.fields.n000.20220920.t15z.nc",
        "nos.creofs.fields.n001.20220920.t15z.nc",
        "nos.creofs.fields.n002.20220920.t15z.nc",
        "nos.creofs.fields.n003.20220920.t15z.nc",
        "nos.creofs.fields.n004.20220920.t15z.nc",
        "nos.creofs.fields.n005.20220920.t15z.nc",
        "nos.creofs.fields.n006.20220920.t15z.nc",
    ]

    df = mc.filedates2df(fnames)

    # also no duplicate in ordered_fnames
    ordered_fnames = [
        "nos.creofs.fields.n000.20220920.t15z.nc",
        "nos.creofs.fields.n001.20220920.t15z.nc",
        "nos.creofs.fields.n002.20220920.t15z.nc",
        "nos.creofs.fields.n003.20220920.t15z.nc",
        "nos.creofs.fields.n004.20220920.t15z.nc",
        "nos.creofs.fields.n005.20220920.t15z.nc",
        "nos.creofs.fields.f000.20220920.t15z.nc",
        "nos.creofs.fields.f001.20220920.t15z.nc",
        "nos.creofs.fields.f002.20220920.t15z.nc",
        "nos.creofs.fields.f003.20220920.t15z.nc",
    ]

    assert ordered_fnames == list(df["filenames"].values)


# @pytest.mark.slow
def test_urlpath_after_select():
    """urlpath is replaced after select_date_range"""

    main_cat = mc.setup(boundaries=False)

    day = "2022-1-1"
    source = mc.select_date_range(
        main_cat["CREOFS"]["ncei-archive-noagg"],
        start_date=day,
        end_date=day,
        override=True,
    )

    # compare datetimes instead of filenames since filenames are not unique
    known_dates = [
        pd.Timestamp("2022-01-01 00:00:00"),
        pd.Timestamp("2022-01-01 01:00:00"),
        pd.Timestamp("2022-01-01 02:00:00"),
        pd.Timestamp("2022-01-01 03:00:00"),
        pd.Timestamp("2022-01-01 04:00:00"),
        pd.Timestamp("2022-01-01 05:00:00"),
        pd.Timestamp("2022-01-01 06:00:00"),
        pd.Timestamp("2022-01-01 07:00:00"),
        pd.Timestamp("2022-01-01 08:00:00"),
        pd.Timestamp("2022-01-01 09:00:00"),
        pd.Timestamp("2022-01-01 10:00:00"),
        pd.Timestamp("2022-01-01 11:00:00"),
        pd.Timestamp("2022-01-01 12:00:00"),
        pd.Timestamp("2022-01-01 13:00:00"),
        pd.Timestamp("2022-01-01 14:00:00"),
        pd.Timestamp("2022-01-01 15:00:00"),
        pd.Timestamp("2022-01-01 16:00:00"),
        pd.Timestamp("2022-01-01 17:00:00"),
        pd.Timestamp("2022-01-01 18:00:00"),
        pd.Timestamp("2022-01-01 19:00:00"),
        pd.Timestamp("2022-01-01 20:00:00"),
        pd.Timestamp("2022-01-01 21:00:00"),
        pd.Timestamp("2022-01-01 22:00:00"),
        pd.Timestamp("2022-01-01 23:00:00"),
    ]

    assert source.dates == known_dates


def test_setting_std_name():
    """Test updating standard name metadata attributes."""
    lon = xr.DataArray(
        data=np.arange(360), dims=("lon",), attrs={"units": "degrees_east"}
    )
    lat = xr.DataArray(
        data=np.arange(-90, 90), dims=("lat",), attrs={"units": "degrees_north"}
    )
    temp = xr.DataArray(
        data=np.arange(360 * 180).reshape(180, 360),
        dims=("lat", "lon"),
        attrs={"units": "degrees_C"},
    )
    ds = xr.Dataset({"lon": lon, "lat": lat, "temp": temp})
    metadata = {
        "standard_names": {
            "longitude": ["lon"],
            "latitude": ["lat"],
        }
    }
    ds = process.add_attributes(ds, metadata=metadata)
    assert ds["lon"].attrs["standard_name"] == "longitude"
    assert ds["lat"].attrs["standard_name"] == "latitude"


@mock.patch("xarray.open_dataset")
@mock.patch("xarray.open_mfdataset")
def test_wrong_time_range(mock_open_dataset, mock_open_mfdataset):
    ds = xr.Dataset()
    dates = pd.date_range("2000-1-1", "2000-2-1", freq="1D")
    ds["time"] = ("time", dates, {"axis": "T"})
    mock_open_dataset.return_value = ds
    mock_open_mfdataset.return_value = ds

    # have to use a real cat/source pair to get this to work, but it isn't actually called in to_dask
    main_cat = mc.setup(boundaries=False)
    cat0 = main_cat[list(main_cat)[0]]
    source0 = cat0[list(cat0)[0]]
    source0.metadata = {}
    source0 = mc.select_date_range(
        source0, start_date="2001-10-1", end_date="2001-10-2"
    )
    with pytest.raises(RuntimeError):
        source0.to_dask()



"""
This file contains all information for transforming the Datasets.
"""
import warnings

from typing import Optional

import cf_xarray  # noqa
import numpy as np
import pandas as pd
import xarray as xr

from intake.source.derived import GenericTransform

import model_catalogs as mc


# extract_model might be necessary for reading in model output, if using FVCOM model
try:
    import extract_model  # noqa: F401

    EM_AVAILABLE = True
except ImportError:  # pragma: no cover
    EM_AVAILABLE = False  # pragma: no cover


yesterday = pd.Timestamp.today() - pd.Timedelta("1 day")


class DatasetTransform(GenericTransform):
    """Transform where the input and output are both Dask-compatible Datasets

    This derives from GenericTransform, and you must supply ``transform`` and
    any ``transform_kwargs``.
    """

    input_container = "xarray"
    container = "xarray"
    optional_params = {}
    _ds = None

    @property
    def urlpath(self):
        """Data location for target

        Can be overwritten by `update_urlpath`

        Returns
        -------
        list
            Location(s) for where data can be found
        """

        if not hasattr(self, "_urlpath"):
            self.target
        return self._urlpath

    @property
    def status(self):
        """Status of server for source.

        Returns
        -------
        bool
            If True, server was reachable.
        """

        if not hasattr(self, "_status"):

            if self.target.describe()["driver"][0] == "opendap":
                suffix = ".das"
            else:
                suffix = ""
            self._status = mc.status(mc.astype(self.urlpath, list)[0], suffix=suffix)
        return self._status

    @property
    def dates(self):
        """Dates associated with urlpath files

        ...if there is more than one. Doesn't work for static links or RTOFS models. So, this is really for NOAA OFS models.

        Returns
        -------
        list
            Ordered dates to match `urlpath` locations.
        """

        if "catloc" in self.metadata:
            dates = []
            for url in self.urlpath:
                dates.extend(mc.astype(mc.file2dt(url), list))
            self._dates = dates
        else:
            self._dates = None

        return self._dates

    @property
    def target(self):
        """Connect target into Transform

        This way can expose some information to query. This will only run once per object.

        Returns
        -------
        Intake Source
            Source that is the target of the object Transform
        """

        # all functions here call follow_target to start
        # need to pick the source only once
        if self._source is None:
            self._pick()

            # if "yesterday" is in user_parameters for original source
            # run with that sent in
            if any(
                [
                    "yesterday" in d.values()
                    for d in self._source.describe()["user_parameters"]
                ]
            ):
                self._source = self._source(yesterday=yesterday)

            # self is of type DatasetTransform instead of OpenDapSource
            # since the OpenDapSource is the target of the Transform
            # but make it easy to check urlpath
            self._urlpath = self._source.urlpath

        return self._source

    def update_urlpath(self):
        """Update urlpath for transform.

        Run this in `select_date_range` for aggregated sources. This can be run more than once.
        """

        if not hasattr(self, "target"):
            self.target

        kwargs = self._params["transform_kwargs"]

        # this is the one that is used when .to_dask() is run
        self._source.urlpath = kwargs["urlpath"]

        # but this one is printed when looking at the source, so better change it too
        self._source._captured_init_kwargs["urlpath"] = kwargs["urlpath"]

        # self is of type DatasetTransform instead of OpenDapSource
        # since the OpenDapSource is the target of the Transform
        # but make it easy to check urlpath
        self._urlpath = kwargs["urlpath"]

    def to_dask(self):
        """Makes it so can read in model output.

        Returns
        -------
        Dataset
            xarray Dataset that has been read in
        """

        if self._ds is None:

            # This is a special flag to be able to just quickly connect to the model output
            # without all the extra checks and processing (other than to transform the original
            # target). Used to find_availability without modifying the Dataset itself.
            if (
                "skip_dask_processing" in self._kwargs
                and self._kwargs["skip_dask_processing"]
            ):
                return self._transform(
                    self._source.to_dask(),
                    metadata=self.metadata,
                )

            if not hasattr(self, "target"):
                self.target

            kwargs = self._params["transform_kwargs"]

            # Checks to catch potential user pitfalls
            # Make sure that user has filled in urlpath if needed if model_source is coops-forecast-noagg.
            # These are unaggregated and come with 2 sample files but user should be warned if they haven't been
            # replaced since it might be a mistake
            if "sample_locs" in self._source.metadata:
                if self._source.urlpath == self._source.metadata["sample_locs"]:
                    warnings.warn(
                        "Note that you are using the original example files in your input source. You may want to instead first run `mc.select_date_range()` to search for and add the correct model output files for your desired date range, then run `to_dask()`.",  # noqa: E501
                        UserWarning,
                    )

            # Make sure that source has urlpath
            # check for if the urlpath is null and if so `select_date_range()`
            # needs to be run to fill it in
            elif self.target.urlpath is None:
                raise KeyError(
                    "The input source `urlpath` does not have a value. You probably want to run `mc.select_date_range()` before running `to_dask()`."  # noqa: E501
                )

            # Alert if triangularmesh engine is required (from FVCOM) but not present
            if (
                self.target.describe()["driver"][0] == "opendap"
                and self.target.engine == "triangularmesh_netcdf"
                and not EM_AVAILABLE
            ):
                raise ModuleNotFoundError(  # pragma: no cover
                    "`extract_model` is not available but contains the 'triangularmesh_netcdf' engine that is required for a model."
                )

            # if "yesterday" is in user_parameters for original source, check that yesterday is still
            # yesterday. Otherwise, info in source is old.
            if any(
                [
                    "yesterday" in d.values()
                    for d in self._source.describe()["user_parameters"]
                ]
            ):
                check_yesterday = pd.Timestamp.today() - pd.Timedelta("1 day")
                if yesterday.date() != check_yesterday.date():
                    warnings.warn(
                        f"You may be running with an out of date source, and you may consider restarting the kernel to update. Yesterday from code: {yesterday.date()}, yesterday right now: {check_yesterday.date()}.",  # noqa: E501
                        UserWarning,
                    )

            # if "today" is in user_parameters for original source, check that today is still
            # today. Otherwise, info in source is old.
            if any(
                [
                    "tod" in d.values()
                    for d in self._source.describe()["user_parameters"]
                ]
            ):
                check_today = pd.Timestamp.today()
                today = [
                    d["default"]
                    for d in self._source.describe()["user_parameters"]
                    if "tod" in d.values()
                ][0]
                if today.date() != check_today.date():
                    warnings.warn(
                        f"You may be running with an out of date source, and you may consider restarting the kernel to update. Today from code: {today.date()}, today right now: {check_today.date()}.",  # noqa: E501
                        RuntimeWarning,
                    )

            # This sends the metadata to `add_attributes()`
            self._ds = self._transform(
                self._source.to_dask(),
                metadata=self.metadata,
            )

            # drop any time duplicates that may be present (RTOFS can have)
            self._ds = self._ds.drop_duplicates(dim=self._ds.cf.axes["T"])

            # check for 'urlpath' update being sent in, if so use it to
            # subselect ds in time
            if "start_date" in kwargs and "end_date" in kwargs:

                try:
                    ds_temp = self._ds.cf.sel(
                        T=slice(kwargs["start_date"], kwargs["end_date"])
                    )

                    if len(ds_temp.cf["T"]) == 0:
                        self._ds = None
                        raise RuntimeError(
                            f"The time slice requested for source {self.name}, {self.cat.name}, {kwargs['start_date']} to {kwargs['end_date']}, results in no times in the Dataset."
                        )
                    else:
                        self._ds = ds_temp

                except KeyError:
                    # self._ds = self._ds
                    warnings.warn(
                        f"The time slice requested for source {self.name}, {self.cat.name}, {kwargs['start_date']} to {kwargs['end_date']}, did not result in a valid Dataset, and so was not used.",
                        RuntimeWarning,
                    )

        return self._ds

    def read(self):
        """Same here."""
        return self.to_dask()


def add_attributes(ds, metadata: Optional[dict] = None):
    """Update Dataset metadata.

    Update the Dataset metadata with metadata passed in from catalog files.

    Parameters
    ----------
    ds : Dataset
        xarray Dataset containing model output.
    metadata : dict, optional
        Metadata that has processing information to apply

    Returns
    -------
    Dataset
        Improved Dataset.
    """
    # set standard_names for all variables
    if metadata is not None and "standard_names" in metadata:
        for stan_name, var_names in metadata["standard_names"].items():
            if not isinstance(var_names, list):
                var_names = [var_names]
            for var_name in var_names:
                if var_name in ds.variables:
                    ds[var_name].attrs["standard_name"] = stan_name

    # # Run code to find vertical coordinates
    # try:
    #     # create name mapping
    #     snames = ['ocean_s_coordinate_g1', 'ocean_s_coordinate_g2', 'ocean_sigma_coordinate']
    #     s_vars = [standard_names[sname] for sname in snames if sname in standard_names][0]
    #     z_vars = axis['Z']
    #     outnames = {s_var: z_var for s_var, z_var in zip(s_vars, z_vars)}
    #     ds.cf.decode_vertical_coords(outnames=outnames)
    # except Exception:
    #     pass

    if metadata is not None and "coords" in metadata:
        ds = ds.assign_coords({k: ds[k] for k in metadata["coords"]})

    # set axis attributes (T, X, Y, Z potentially)
    if metadata is not None and "axis" in metadata:
        for ax_name, var_names in metadata["axis"].items():
            var_names = mc.astype(var_names, list)
            for var_name in var_names:

                # Check dims, coords, and data_vars:
                if (
                    var_name in ds.dims
                    or var_name in ds.data_vars.keys()
                    or var_name in ds.coords
                ):
                    # var_name needs to be a coord to have attributes
                    if var_name not in ds.coords:
                        ds = ds.assign_coords(
                            {
                                var_name: (
                                    var_name,
                                    np.arange(ds[var_name].size),
                                    {"axis": ax_name},
                                )
                            }
                        )
                    else:
                        ds[var_name].attrs["axis"] = ax_name

                else:
                    warnings.warn(
                        f"The variable {var_name} input in a catalog file is not present in the Dataset.",
                        UserWarning,
                    )

    # this won't run for e.g. GFS which has multiple time variables
    # but also doesn't need to have the calendar updated
    try:
        attrs = ds[ds.cf["T"].name].attrs
        if ("calendar" in attrs) and (attrs["calendar"] == "gregorian_proleptic"):
            attrs["calendar"] = "proleptic_gregorian"
            ds[ds.cf["T"].name].attrs = attrs
    except KeyError:
        pass

    # check for a dataset having both `missing_value` and `_FillValue` attributes defined
    # delete `missing_value` in this case (arbitrary decision of which to delete)
    for data_var in ds.data_vars:
        if "missing_value" in ds[data_var].attrs and "_FillValue" in ds[data_var].attrs:
            del ds[data_var].attrs["missing_value"]

    # decode times if times are floats.
    # Some datasets like GFS have multiple time coordinates for different phenomena like
    # precipitation accumulation vs winds vs surface albedo average.
    if (
        metadata is not None
        and "axis" in metadata
        and "T" in metadata["axis"]
        and isinstance(metadata["axis"]["T"], list)
        and len(metadata["axis"]["T"]) > 1
    ):
        for time_var in metadata["axis"]["T"]:
            if ds[time_var].dtype == "float64":
                ds = xr.decode_cf(ds, decode_times=True)
                break
    elif "T" in ds.cf and ds.cf["T"].dtype == "float64":
        ds = xr.decode_cf(ds, decode_times=True)

    if metadata is not None and "formula_terms" in metadata:
        for varname in metadata["formula_terms"]:
            ds[varname].attrs["formula_terms"] = metadata["formula_terms"][varname]

    # This is an internal attribute used by netCDF which xarray doesn't know or care about, but can
    # be returned from THREDDS.
    if "_NCProperties" in ds.attrs:
        del ds.attrs["_NCProperties"]

    return ds



"""Test different ways to input catalogs."""

from pathlib import PurePath

import intake
import numpy as np
import xarray as xr

import model_catalogs as mc


def test_default_catalogs_intake():
    """Make sure some default 'mc_' catalogs are present since are required."""

    assert "mc_CBOFS" in list(intake.cat)
    assert "mc_CIOFS" in list(intake.cat)


def test_setup_default():
    """Check that default `mc_` catalogs come through setup."""

    main_cat = mc.setup(boundaries=False)
    assert "CBOFS" in list(main_cat)
    assert "CIOFS" in list(main_cat)


def test_open_catalog(tmpdir):
    dsfname = PurePath(tmpdir) / "example_dataset.nc"
    fname = PurePath(tmpdir) / "example_catalog.yaml"

    ds = xr.Dataset()
    ds["time"] = (
        "Time",
        np.arange(10),
        # {"standard_name": "time"},
    )
    ds["Lat"] = (
        "y",
        np.arange(10),
        # {"units": "degrees_north", "standard_name": "latitude"},
    )
    ds["Lon"] = (
        "x",
        np.arange(10),
        # {"units": "degrees_east", "standard_name": "longitude"},
    )
    ds.to_netcdf(dsfname)

    catalog_text = f"""
    name: example
    metadata:
        alpha_shape: [4,50]  # dd, alpha
    sources:
        example_source:
            args:
                urlpath: {dsfname}
            description: ''
            driver: netcdf
            metadata:
                axis:
                    T: Time
                    X: x
                    Y: y
                standard_names:
                    time: time
                    latitude: Lat
                    longitude: Lon
                coords:
                    - Lon
                    - Lat
                    - Time
    """

    with open(fname, "w") as fp:
        fp.write(catalog_text)
    cat = mc.open_catalog(fname)
    ds = cat["example_source"].to_dask()
    assert ds.cf.axes["X"] == ["x"]
    assert ds.cf.axes["Y"] == ["y"]
    assert ds.cf.axes["T"] == ["Time"]
    assert ds.cf.coordinates["latitude"] == ["Lat"]
    assert ds.cf.coordinates["longitude"] == ["Lon"]
    assert ds.cf.coordinates["time"] == ["Time"]
